/* Elvira Bluetooth FTP 
 * Copyright (C) 2007  Alberto Salas Bragado
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "btsearch.h"

#define ObexFTP 0
#define ObexPUSH 1

int get_channel_from_pd(sdp_data_t *value);
int get_ftp_port(char *bt_address);


/* http://people.csail.mit.edu/albert/bluez-intro/c404.html */
int hci_search(device_t **list) {
	
	//Puntero a estructura con informacion de un dispositivo, resultado de una busqueda
	inquiry_info *ii= NULL;
	
	//Duración de la búsqueda de dispositivos = len * 1.28 segundos
	int len = 5;
	
	//Maximo numero de dispositivos a devolver por BlueZ y numero de dispositivos encontrados
	int max_rsp = 256, num_rsp;
	
	//Numero de dispositivos devueltos por esta funcion
	//(Sólo los que exporten FTP)
	int num_ftp=0;
	
	//Identificador de dispositivo y socket del microcontrolador
	int dev_id, sock;
	
	//Flags para la busqueda de dispositivos
	//IREQ_CACHE_FLUSH obliga a borrar los dispositivos de la cache de antiguas búsquedas
	//antes de iniciar una nueva.
	int flags = IREQ_CACHE_FLUSH;
	
	//Resto de variables necesarias
	int i;
	device_t *lista = NULL;
	char mac_aux[100];
	int port_aux;

	//A los dispositivos Bluetooth se les asigna un número entero identificativo, empezando desde 0.
	//Para usar un dispositivo es necesario solicitarlo usando su número identificativo.
	//Como sólo suele haber un dispositivo en el sistema, con NULL indicamos que queremos usar el primer
	//dispositivo disponible. (Nota: no es buena idea usar un 0 en esta funcion sin comprobar si esta disponible).
	dev_id = hci_get_route(NULL);
	
	//Solicitamos la conexión con el microcontrolador del dispositivo a usar, NO con un dispositivo remoto.
	//Es necesario para algunas operaciones.
	sock = hci_open_dev(dev_id);
	
	//Comprobamos si hemos obtenenido un dispositivo y socket.
	if (dev_id < 0 || sock < 0) {
		/*perror("Accediendo al dispositivo local");
		exit(1);*/
		return E_LOCAL_DISP;
	}

	//Lanzamos la busqueda de dispositivos que nos devolvera una lista de dispositivos encontrados en ii
	//y el numero de ellos en num_rsp
	num_rsp = hci_inquiry(dev_id, len, max_rsp, NULL, &ii, flags);
	//Comprobamos que la búsqueda se haya hecho con éxito
	if (num_rsp < 0) {
		//Liberar memoria usada en la busqueda de dispositivos
		free(ii);
		//Cerrar el socket con el microcontrolador
		hci_close_dev(sock);
		//Devolver puntero nulo como lista de dispositivos
		*list = NULL;
		//Devolver el error
		return E_SEARCH_DISP;
	}	

	//Recorremos los dispositivos encontrados
	for (i = 0; i < num_rsp; i++) {
		
		//Extrar la mac y puerto del dispositivo
		ba2str(&(ii[i].bdaddr), mac_aux);
		port_aux = get_ftp_port(mac_aux);
		
		//Si el dispositivo exporta el servicio FTP lo añadimos a la lista
		if (port_aux > 0) {
		
			//Aumentamos el numero de dispositivos que exportan FTP
			num_ftp++;
			
			//Reservamos memoria
			lista = (device_t *) realloc(lista, sizeof(device_t)*num_ftp);
			
			//Obtenemos el nombre corto
			if (hci_read_remote_name(sock, &(ii[i].bdaddr), 255, lista[num_ftp-1].name, 0) < 0)
				strcpy(lista[num_ftp-1].name, "[sin nombre]");
			
			//Asignamos mac y puerto obtenidos antes
			strcpy(lista[num_ftp-1].mac, mac_aux);
			lista[num_ftp-1].port = port_aux;
			
		}
		
	}

	//Liberar memoria usada en la busqueda de dispositivos
	free(ii);
	
	//Cerrar el socket con el microcontrolador
	hci_close_dev(sock);
	
	//Devolver lista de dispositivos
	*list = lista;
	
	return num_ftp;
}

//Funcion que obtiene el canal RFCOMM utilizado por el perfil OBEX Object Push en el dispositivo
int get_ftp_port(char *bt_address) {

	//direccion mac del dispositivo
	bdaddr_t bdaddr;
	//uuid del servicio OBEX FTP
    uint16_t class = 0x1106; 
	//nodos de listas enlazadas de BlueZ
    sdp_list_t *attrid, *search, *seq, *next;
	//mascara para extraer las propiedades de un servicio del servidor SDP
    uint32_t range = 0x0000ffff;
	//indentificador de conexión con el servidor SDP
    sdp_session_t *sess;
	//valor que devolveremos, canal del servicio buscado
    uint32_t channel = -1;
	//estructura que identificara al servicio buscado
    uuid_t svc_id;
	
	//crear la estructura con la direccion del dispositivo
    str2ba(bt_address, &bdaddr);
	
	// conectar al servidor SDP del dispositivo remoto
    sess = sdp_connect(BDADDR_ANY, &bdaddr, SDP_RETRY_IF_BUSY);
	 
	//Comprobamos si ha habido exito en la conexion
    if (!sess) {
    	return -1;
    }
 
    //construir el UUID del servicio que estamos buscando
    sdp_uuid16_create(&svc_id, class);
	//specificamos que queremos una lista de todas las propiedades del rango marcado
    attrid = sdp_list_append(0, &range);
	//especificamos el servicio que estamos buscando
    search = sdp_list_append(0, &svc_id);
	 
	//solicitamos una lista de servicios con las caracteristicas dadas
    if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
    	sdp_close(sess);
    	return -1;
    }
	 
	//liberamos memoria innecesaria
    sdp_list_free(attrid, 0);
    sdp_list_free(search, 0);
 
    //recorremos la lista de service records obtenida por el servidor SDP
    for (; seq; seq = next) {
    	
    	//datos del nodo actual de la lista
    	sdp_record_t *rec = (sdp_record_t *) seq->data;
	 
    	//obtener la secuencia de protocolos
    	sdp_list_t *proto = 0;
    	if (sdp_get_access_protos(rec, &proto) == 0) {
	 
    		//recorremos la secuencia devuelta de protocolos
    		sdp_list_t* ptr = proto;
    		for(;ptr != NULL;ptr = ptr->next){
		 
    			//obtenemos los datos del nodo de la lista
    			sdp_list_t *protDescSeq = (sdp_list_t *)ptr->data;
    			//recorremos los atributos del protocolo y buscamos el puerto entre ellos				 
    			for(;protDescSeq != NULL;protDescSeq = protDescSeq->next){
			 
    				//usamos otra funcion para extraer el puerto del servicio
    				channel = get_channel_from_pd(protDescSeq->data);
                    
    				//cuando encontramos el puerto salimos
    				if(channel != -1) break;
				 
    			} 
    		} 
    		//liberar memoria
    		sdp_list_free(proto, 0);
    	}
 
    	//siguiente nodo de la lista
    	next = seq->next;
    	//liberar memoria
    	free(seq);
    	//liberar memoria
    	sdp_record_free(rec);
	}
 
	//cerramos la conexión con el servidor SDP remoto
	sdp_close(sess);
	//devolvemos el valor del puerto
	return channel;
}

// Función para obtener el puerto a partir de los atributos de servicio
// Si los atributos no contienen el puerto, devuelve -1
int get_channel_from_pd(sdp_data_t *value) {
	
	int proto = 0;
     
	for (; value; value = value->next) {
		switch (value->dtd) {
			case SDP_UUID16:
			case SDP_UUID32:
			case SDP_UUID128:
				proto = sdp_uuid_to_proto(&value->val.uuid);
				break;
			case SDP_UINT8:
				if (proto == RFCOMM_UUID)
					return value->val.uint8;
				break;
		}
	}
	
	return -1;
	
}
